
// ========================================================================
// Copyright (c) 1996 Mort Bay Consulting Pty. Ltd. All rights reserved.
// $Id: MultiPartRequest.java,v 1.1 2008/03/28 14:40:27 pjr Exp $
// ------------------------------------------------------------------------
// Modified By PJR 1060 Research to support HttpRequest

package org.netkernel.http.util;

import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Hashtable;
import java.util.StringTokenizer;

import javax.servlet.http.HttpServletRequest;

import org.eclipse.jetty.util.StringUtil;

/* ------------------------------------------------------------ */
/** Multipart Form Data request.
 * <p>
 * This class decodes the multipart/form-data stream sent by
 * a HTML form that uses a file input item.
 *
 * <p><h4>Usage</h4>
 * Each part of the form data is named from the HTML form and
 * is available either via getString(name) or getInputStream(name).
 * Furthermore the MIME parameters and filename can be requested for
 * each part.
 * <pre>
 * </pre>
 * 
 * modified by Dong Liu for porting to jetty 6
 * @version $Id: MultiPartRequest.java,v 1.1 2008/03/28 14:40:27 pjr Exp $ Dec 3, 2007 11:50:51 AM
 * @author  Greg Wilkins
 * @author  Jim Crossley 
 * @author Dong Liu
 */
public class MultiPartRequest
{
    /* ------------------------------------------------------------ */
    HttpServletRequest _request;
    LineInput _in;
    String _boundary;
    byte[] _byteBoundary;
    Hashtable _partMap = new Hashtable(10);
    int _char=-2;
    boolean _lastPart=false;
    
    /* ------------------------------------------------------------ */
    /** Constructor. 
     * @param request The request containing a multipart/form-data
     * request
     * @exception IOException IOException
     */
    public MultiPartRequest(HttpServletRequest request)
        throws IOException
    {
        _request=request;
		/*
		String content_type = request.getMimeType();
        if (!content_type.startsWith("multipart/form-data"))
            throw new IOException("Not multipart/form-data request");

        Code.debug("Multipart content type = ",content_type);
        */
        _in = new LineInput(request.getInputStream());
        
		// Extract boundary string
        _boundary=_in.readLine();
		/*
		"--"+
            value(content_type.substring(content_type.indexOf("boundary=")));
        */
        //Code.debug("Boundary=",_boundary);
        _byteBoundary= (_boundary+"--").getBytes(StringUtil.__ISO_8859_1);
       
        loadAllParts();
    }
    
    public void cleanup()
    {  	for (Object part : _partMap.values())
    	{	File f=((Part)part)._file;
    		if (f!=null)
    		{	f.delete();
    		}
    	}
    }
    
    /* ------------------------------------------------------------ */
    /** Get the part names.
     * @return an array of part names
     */
    public String[] getPartNames()
    {
        return (String[]) _partMap.keySet().toArray(new String[0]);
    }
    
    /* ------------------------------------------------------------ */
    /** Check if a named part is present 
     * @param name The part
     * @return true if it was included 
     */
    public boolean contains(String name)
    {
        Part part = (Part)_partMap.get(name);
        return (part!=null);
    }
    
    /* ------------------------------------------------------------ */
    /** Get the data of a part as a string.
     * @param name The part name 
     * @return The part data
     */
    public String getString(String name)
    {
        Part part = (Part)_partMap.get(name);
        if (part==null)
            return null;
        if (part._data!=null)
        {	return new String(part._data);
        }
        else
        {	return "";
        }
    }
    

    /* ------------------------------------------------------------ */
    /** Get the data of a part as a stream.
     * @param name The part name 
     * @return Stream providing the part data
     */
    public InputStream getInputStream(String name) throws FileNotFoundException
    {
        Part part = (Part)_partMap.get(name);
        if (part==null)
            return null;
        if (part._data!=null)
        {	return new ByteArrayInputStream(part._data);
        }
        else
        {	return new FileInputStream(part._file);
        }
    }
    
    /* ------------------------------------------------------------ */
    /** Get the MIME parameters associated with a part.
     * @param name The part name 
     * @return Hashtable of parameters
     */
    public Hashtable getParams(String name)
    {
        Part part = (Part)_partMap.get(name);
        if (part==null)
            return null;
        return part._headers;
    }
    
    /* ------------------------------------------------------------ */
    /** Get any file name associated with a part.
     * @param name The part name 
     * @return The filename
     */
    public String getFilename(String name)
    {
        Part part = (Part)_partMap.get(name);
        if (part==null)
            return null;
        return part._filename;
    }
    
	/* ------------------------------------------------------------ */
    /** Get content length associated with a part.
     * @param name The part name 
     * @return content length
     */
    public int getContentLength(String name)
    {
        Part part = (Part)_partMap.get(name);
        if (part==null)
            return -1;
        
        if (part._data!=null)
        {	return part._data.length;
        }
        else
        {	return (int)part._file.length();
        }
    }
	
    
    /* ------------------------------------------------------------ */
    private void loadAllParts()
        throws IOException
    {
        // Get first boundary
		String line;
        /*
		String line = _in.readLine();
        if (!line.equals(_boundary))
        {
            Code.warning(line);
            throw new IOException("Missing initial multi part boundary");
        }
        */
        // Read each part
        while (!_lastPart)
        {
            // Read Part headers
            Part part = new Part();
            
            String content_disposition=null;
            while ((line=_in.readLine())!=null)
            {
                // If blank line, end of part headers
                if (line.length()==0)
                    break;

                //Code.debug("LINE=",line);
                
                // place part header key and value in map
                int c = line.indexOf(':',0);
                if (c>0)
                {
                    String key = line.substring(0,c).trim().toLowerCase();
                    String value = line.substring(c+1,line.length()).trim();
                    String ev = (String) part._headers.get(key);
                    part._headers.put(key,(ev!=null)?(ev+';'+value):value);
                    //Code.debug(key,": ",value);
                    if (key.equals("content-disposition"))
                        content_disposition=value;
                }
            }

            // Extract content-disposition
            boolean form_data=false;
            if (content_disposition==null)
            {
                throw new IOException("Missing content-disposition");
            }
            
            StringTokenizer tok =
                new StringTokenizer(content_disposition,";");
            while (tok.hasMoreTokens())
            {
                String t = tok.nextToken().trim();
                String tl = t.toLowerCase();
                if (t.startsWith("form-data"))
                    form_data=true;
                else if (tl.startsWith("name="))
                    part._name=value(t);
                else if (tl.startsWith("filename="))
                    part._filename=value(t);
            }

            // Check disposition
            if (!form_data)
            {
                //Code.warning("Non form-data part in multipart/form-data");
                continue;
            }
            if (part._name==null || part._name.length()==0)
            {
                //Code.warning("Part with no name in multipart/form-data");
                continue;
            }
            //Code.debug("name=",part._name);
            //Code.debug("filename=",part._filename);
            _partMap.put(part._name,part);
            readBytes(part,100000);
        }       
    }

    /* ------------------------------------------------------------ */
    private void readBytes(Part part, int aInMemoryThreshold)
        throws IOException
    {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        File f=null;
        OutputStream fos=null;
        OutputStream os=baos;
        
        int c;
        boolean cr=false;
        boolean lf=false;
        
        // loop for all lines`
        while (true)
        {
            int b=0;
            while ((c=(_char!=-2)?_char:_in.read())!=-1)
            {
                _char=-2;

                // look for CR and/or LF
                if (c==13 || c==10)
                {
                    if (c==13) _char=_in.read();
                    break;
                }

                // look for boundary
                if (b>=0 && b<_byteBoundary.length && c==_byteBoundary[b])
                    b++;
                else
                {
                    // this is not a boundary
                    if (cr) os.write(13);
                    if (lf) os.write(10);
                    cr=lf=false;
                    
                    if (b>0)
                        os.write(_byteBoundary,0,b);
                    b=-1;
                  
                    os.write(c);
                }
            }

            // check partial boundary
             if ((b>0 && b<_byteBoundary.length-2) ||
                 (b==_byteBoundary.length-1))
             {
                 if (cr) os.write(13);
                 if (lf) os.write(10);
                 cr=lf=false;
                 os.write(_byteBoundary,0,b);
                 b=-1;
             }
            
            // boundary match
            if (b>0 || c==-1)
            {
                if (b==_byteBoundary.length)
                    _lastPart=true;
                if (_char==10) _char=-2;
                break;
            }
            
            // handle CR LF
            if (cr) os.write(13);
            if (lf) os.write(10);
            cr=(c==13);
            lf=(c==10 || _char==10);
            if (_char==10) _char=-2;  
            
            if (os==baos && baos.size()>aInMemoryThreshold)
            {
            	//switch to file
            	f=File.createTempFile("MultiPartRequest", null);
            	fos=new BufferedOutputStream(new FileOutputStream(f));
            	baos.writeTo(fos);
            	baos=null;
            	os=fos;
            }
        }
        
        if (os==baos)
        {	part._data=baos.toByteArray();
        }
        else
        {	fos.flush();
        	fos.close();
        	part._file=f;
        }
    }
    
    
    /* ------------------------------------------------------------ */
    private String value(String nameEqualsValue)
    {   
        String value =
            nameEqualsValue.substring(nameEqualsValue.indexOf('=')+1).trim();
        
        int i=value.indexOf(';');
        if (i>0)
            value=value.substring(0,i);
        if (value.startsWith("\""))
        {
            value=value.substring(1,value.indexOf('"',1));
        }
        
        else
        {
            i=value.indexOf(' ');
            if (i>0)
                value=value.substring(0,i);
        }
        return value;
    }
    
    /* ------------------------------------------------------------ */
    private class Part
    {
        String _name=null;
        String _filename=null;
        Hashtable _headers= new Hashtable(10);
        byte[] _data=null;
        File _file=null;
    }    
};
